/*
 * Copyright (c) 2017 Digital Bazaar, Inc. All rights reserved.
 */
'use strict';

var _require = require('../util');

const parseLinkHeader = _require.parseLinkHeader,
      buildHeaders = _require.buildHeaders,
      promisify = _require.promisify;

var _require2 = require('../constants');

const LINK_HEADER_REL = _require2.LINK_HEADER_REL;

const JsonLdError = require('../JsonLdError');
const RequestQueue = require('../RequestQueue');

/**
 * Creates a built-in node document loader.
 *
 * @param options the options to use:
 *          secure: require all URLs to use HTTPS.
 *          strictSSL: true to require SSL certificates to be valid,
 *            false not to (default: true).
 *          maxRedirects: the maximum number of redirects to permit, none by
 *            default.
 *          request: the object which will make the request, default is
 *            provided by `https://www.npmjs.com/package/request`.
 *          headers: an object (map) of headers which will be passed as request
 *            headers for the requested document. Accept is not allowed.
 *          usePromise: true to use a promises API, false for a
 *            callback-continuation-style API; false by default.
 *
 * @return the node document loader.
 */
module.exports = function () {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  let secure = _ref.secure;
  var _ref$strictSSL = _ref.strictSSL;
  let strictSSL = _ref$strictSSL === undefined ? true : _ref$strictSSL;
  var _ref$maxRedirects = _ref.maxRedirects;
  let maxRedirects = _ref$maxRedirects === undefined ? -1 : _ref$maxRedirects,
      request = _ref.request;
  var _ref$headers = _ref.headers;
  let headers = _ref$headers === undefined ? {} : _ref$headers,
      usePromise = _ref.usePromise;

  headers = buildHeaders(headers);
  request = request || require('request');
  const http = require('http');
  // TODO: disable cache until HTTP caching implemented
  //const cache = new DocumentCache();

  // TODO: simplify w/async
  const queue = new RequestQueue();
  if (usePromise) {
    return queue.wrapLoader(function (url) {
      return promisify(loadDocument, url, []);
    });
  }

  return queue.wrapLoader(function (url, callback) {
    loadDocument(url, [], callback);
  });

  function loadDocument(url, redirects, callback) {
    if (url.indexOf('http:') !== 0 && url.indexOf('https:') !== 0) {
      return callback(new JsonLdError('URL could not be dereferenced; only "http" and "https" URLs are ' + 'supported.', 'jsonld.InvalidUrl', { code: 'loading document failed', url: url }), { contextUrl: null, documentUrl: url, document: null });
    }
    if (secure && url.indexOf('https') !== 0) {
      return callback(new JsonLdError('URL could not be dereferenced; secure mode is enabled and ' + 'the URL\'s scheme is not "https".', 'jsonld.InvalidUrl', { code: 'loading document failed', url: url }), { contextUrl: null, documentUrl: url, document: null });
    }
    // TODO: disable cache until HTTP caching implemented
    let doc = null; //cache.get(url);
    if (doc !== null) {
      return callback(null, doc);
    }

    request({
      url: url,
      headers: headers,
      strictSSL: strictSSL,
      followRedirect: false
    }, handleResponse);

    function handleResponse(err, res, body) {
      doc = { contextUrl: null, documentUrl: url, document: body || null };

      // handle error
      if (err) {
        return callback(new JsonLdError('URL could not be dereferenced, an error occurred.', 'jsonld.LoadDocumentError', { code: 'loading document failed', url: url, cause: err }), doc);
      }
      const statusText = http.STATUS_CODES[res.statusCode];
      if (res.statusCode >= 400) {
        return callback(new JsonLdError('URL could not be dereferenced: ' + statusText, 'jsonld.InvalidUrl', {
          code: 'loading document failed',
          url: url,
          httpStatusCode: res.statusCode
        }), doc);
      }

      // handle Link Header
      if (res.headers.link && res.headers['content-type'] !== 'application/ld+json') {
        // only 1 related link header permitted
        const linkHeader = parseLinkHeader(res.headers.link)[LINK_HEADER_REL];
        if (Array.isArray(linkHeader)) {
          return callback(new JsonLdError('URL could not be dereferenced, it has more than one associated ' + 'HTTP Link Header.', 'jsonld.InvalidUrl', { code: 'multiple context link headers', url: url }), doc);
        }
        if (linkHeader) {
          doc.contextUrl = linkHeader.target;
        }
      }

      // handle redirect
      if (res.statusCode >= 300 && res.statusCode < 400 && res.headers.location) {
        if (redirects.length === maxRedirects) {
          return callback(new JsonLdError('URL could not be dereferenced; there were too many redirects.', 'jsonld.TooManyRedirects', {
            code: 'loading document failed',
            url: url,
            httpStatusCode: res.statusCode,
            redirects: redirects
          }), doc);
        }
        if (redirects.indexOf(url) !== -1) {
          return callback(new JsonLdError('URL could not be dereferenced; infinite redirection was detected.', 'jsonld.InfiniteRedirectDetected', {
            code: 'recursive context inclusion',
            url: url,
            httpStatusCode: res.statusCode,
            redirects: redirects
          }), doc);
        }
        redirects.push(url);
        return loadDocument(res.headers.location, redirects, callback);
      }
      // cache for each redirected URL
      redirects.push(url);
      // TODO: disable cache until HTTP caching implemented
      /*for(let i = 0; i < redirects.length; ++i) {
        cache.set(
          redirects[i],
          {contextUrl: null, documentUrl: redirects[i], document: body});
      }*/
      callback(err, doc);
    }
  }
};